"""
ErrorClass Plugins
------------------
ErrorClass plugins provide an easy way to add support for custom
handling of particular classes of exceptions.
An ErrorClass plugin defines one or more ErrorClasses and how each is
handled and reported on. Each error class is stored in a different
attribute on the result, and reported separately. Each error class must
indicate the exceptions that fall under that class, the label to use
for reporting, and whether exceptions of the class should be
considered as failures for the whole test run.
ErrorClasses use a declarative syntax. Assign an ErrorClass to the
attribute you wish to add to the result object, defining the
exceptions, label and isfailure attributes. For example, to declare an
ErrorClassPlugin that defines TodoErrors (and subclasses of TodoError)
as an error class with the label 'TODO' that is considered a failure,
do this:
    >>> class Todo(Exception):
    ...     pass
    >>> class TodoError(ErrorClassPlugin):
    ...     todo = ErrorClass(Todo, label='TODO', isfailure=True)
The MetaErrorClass metaclass translates the ErrorClass declarations
into the tuples used by the error handling and reporting functions in
the result. This is an internal format and subject to change; you
should always use the declarative syntax for attaching ErrorClasses to
an ErrorClass plugin.
    >>> TodoError.errorClasses # doctest: +ELLIPSIS
    ((<class ...Todo...>, ('todo', 'TODO', True)),)
Let's see the plugin in action. First some boilerplate.
    >>> import sys
    >>> import unittest
    >>> from unittest.runner import _WritelnDecorator
    >>> buf = _WritelnDecorator(sys.stdout)
Now define a test case that raises a Todo.
    >>> class TestTodo(unittest.TestCase):
    ...     def runTest(self):
    ...         raise Todo("I need to test something")
    >>> case = TestTodo()
Prepare the result using our plugin. Normally this happens during the
course of test execution within nose -- you won't be doing this
yourself. For the purposes of this testing document, I'm stepping
through the internal process of nose so you can see what happens at
each step.
    >>> plugin = TodoError()
    >>> from nose.result import _TextTestResult
    >>> result = _TextTestResult(stream=buf, descriptions=0, verbosity=2)
    >>> plugin.prepareTestResult(result)
Now run the test. TODO is printed.
    >>> _ = case(result) # doctest: +ELLIPSIS
    runTest (....TestTodo) ... TODO: I need to test something
Errors and failures are empty, but todo has our test:
    >>> result.errors
    []
    >>> result.failures
    []
    >>> result.todo # doctest: +ELLIPSIS
    [(
        <....TestTodo testMethod=runTest>,
        '...Todo: I need to test something\\n'
    )]
    >>> result.printErrors() # doctest: +ELLIPSIS
    <BLANKLINE>
    ======================================================================
    TODO: runTest (....TestTodo)
    ----------------------------------------------------------------------
    Traceback (most recent call last):
    ...
    ...Todo: I need to test something
    <BLANKLINE>
Since we defined a Todo as a failure, the run was not successful.
    >>> result.wasSuccessful()
    False """
from nose.pyversion import make_instancemethod
from nose.plugins.base import Plugin
from nose.result import TextTestResult
from nose.util import isclass


class MetaErrorClass(type):
    def __init__(self, name, bases, attr):
        errorClasses = []
        for name, detail in list(attr.items()):
            if isinstance(detail, ErrorClass):
                attr.pop(name)
                for cls in detail:
                    errorClasses.append(
                        (cls, (name, detail.label, detail.isfailure)))
        super(MetaErrorClass, self).__init__(name, bases, attr)
        self.errorClasses = tuple(errorClasses)


class ErrorClass(object):
    def __init__(self, *errorClasses, **kw):
        self.errorClasses = errorClasses
        try:
            for key in ('label', 'isfailure'):
                setattr(self, key, kw.pop(key))
        except KeyError:
            raise TypeError(
                "%r is a required named argument for ErrorClass" % key
            )

    def __iter__(self):
        return iter(self.errorClasses)


class ErrorClassPlugin(Plugin, metaclass=MetaErrorClass):
    """Base class for ErrorClass plugins. Subclass this class and declare the
    exceptions that you wish to handle as attributes of the subclass."""
    score = 1000
    errorClasses = ()

    def addError(self, test, err):
        err_cls, a, b = err
        if not isclass(err_cls):
            return None
        classes = [e[0] for e in self.errorClasses]
        if [c for c in classes if issubclass(err_cls, c)]:
            return True
        return None

    def prepareTestResult(self, result):
        if not hasattr(result, 'errorClasses'):
            self.patchResult(result)
        for cls, (storage_attr, label, isfail) in self.errorClasses:
            if cls not in result.errorClasses:
                storage = getattr(result, storage_attr, [])
                setattr(result, storage_attr, storage)
                result.errorClasses[cls] = (storage, label, isfail)

    def patchResult(self, result):
        result.printLabel = print_label_patch(result)
        result._orig_addError, result.addError = \
            result.addError, add_error_patch(result)
        result._orig_wasSuccessful, result.wasSuccessful = \
            result.wasSuccessful, wassuccessful_patch(result)
        if hasattr(result, 'printErrors'):
            result._orig_printErrors, result.printErrors = \
                result.printErrors, print_errors_patch(result)
        if hasattr(result, 'addSkip'):
            result._orig_addSkip, result.addSkip = \
                result.addSkip, add_skip_patch(result)
        result.errorClasses = {}


def add_error_patch(result):
    return make_instancemethod(TextTestResult.addError, result)


def print_errors_patch(result):
    return make_instancemethod(TextTestResult.printErrors, result)


def print_label_patch(result):
    return make_instancemethod(TextTestResult.printLabel, result)


def wassuccessful_patch(result):
    return make_instancemethod(TextTestResult.wasSuccessful, result)


def add_skip_patch(result):
    return make_instancemethod(TextTestResult.addSkip, result)


if __name__ == '__main__':
    import doctest
    doctest.testmod()



"""
Overview
========
The multiprocess plugin enables you to distribute your test run among a set of
worker processes that run tests in parallel. This can speed up CPU-bound test
runs (as long as the number of work processeses is around the number of
processors or cores available), but is mainly useful for IO-bound tests that
spend most of their time waiting for data to arrive from someplace else.
.. _multiprocessing : http://code.google.com/p/python-multiprocessing/

How tests are distributed
=========================
The ideal case would be to dispatch each test to a worker process separately.
This ideal is not attainable in all cases, however, because many test suites
depend on context (class, module, or package) fixtures.
The plugin can't know (unless you tell it -- see below!) if a context fixture
can be called many times concurrently (is re-entrant), or if it can be shared
among tests running in different processes. Therefore, if a context has
fixtures, the default behavior is to dispatch the suite to a worker as a unit.

Controlling distribution
========================
There are two context-level variables to control this default behavior:
* If a context's fixtures are re-entrant, set `_multiprocess_can_split_ = True`
in the context, and the plugin will dispatch tests in suites bound to that
context as if the context had no fixtures. This means that the fixtures will
execute concurrently and multiple times, typically once per test.
* If a context's fixtures can be shared by tests running in different processes
(such as a package-level fixture that starts an external http server or
initializes a shared database) then set ``_multiprocess_shared_ = True`` in
the context. These fixtures will then execute in the primary nose process,
and tests in those contexts will be individually dispatched to run in parallel.

How results are collected and reported
======================================
As each test or suite executes in a worker process, results (failures, errors,
and specially handled exceptions like SkipTest) are collected in that process.
When the worker process finishes, it returns results to the main nose process.
There, any progress output is printed (dots!), and the results from the
test run are combined into a consolidated result set.
When results have been received for all dispatched tests,
or all workers have died, the result summary is output as normal.

Beware!
=======
Not all test suites will benefit from, or even operate correctly using this
plugin. For example, CPU-bound tests will run more slowly if you don't have
multiple processors. There are also some differences in plugin interactions
and behaviors due to the way in which tests are dispatched and loaded.
In general, tests loaded under this plugin operate as if they were always in
directed mode instead of discovered mode. For instance, doctests in test
modules will always be found when using this plugin with the doctest plugin.
But the biggest issue you will face is probably concurrency. Unless you
have kept your tests as religiously pure unit tests, with no side-effects, no
ordering issues, and no external dependencies, chances are you will experience
odd, intermittent and unexplainable failures and errors when using this
plugin. This doesn't necessarily mean the plugin is broken;
it may mean that your test suite is not safe for concurrency."""
import logging
import os
import sys
import time
import unittest
import pickle
import signal
import nose.case
from nose.core import TextTestRunner
from nose import failure
from nose import loader
from nose.plugins.base import Plugin
from nose.pyversion import bytes_
from nose.suite import ContextSuite
from nose.util import test_address
from unittest.runner import _WritelnDecorator
from queue import Empty
from warnings import warn
from io import StringIO

_instantiate_plugins = None
log = logging.getLogger(__name__)
Process = Queue = Pool = Event = Value = Array = None


class TimedOutException(KeyboardInterrupt):

    def __init__(self, value="Timed Out"):
        self.value = value

    def __str__(self):
        return repr(self.value)


def _import_mp():
    global Process, Queue, Pool, Event, Value, Array
    try:
        from multiprocessing import Manager, Process
        old = signal.signal(signal.SIGINT, signal.SIG_IGN)
        m = Manager()
        signal.signal(signal.SIGINT, old)
        Queue, Pool, Event, Value, Array = (
                m.Queue, m.Pool, m.Event, m.Value, m.Array
        )
    except ImportError:
        warn("multiprocessing module is not available, multiprocess plugin "
             "cannot be used", RuntimeWarning)


class TestLet:
    def __init__(self, case):
        try:
            self._id = case.id()
        except AttributeError:
            pass
        self._short_description = case.shortDescription()
        self._str = str(case)

    def id(self):
        return self._id

    def shortDescription(self):
        return self._short_description

    def __str__(self):
        return self._str


class MultiProcess(Plugin):
    """Run tests in multiple processes. Requires processing module."""
    score = 1000
    status = {}

    def options(self, parser, env):
        """Register command-line options."""
        parser.add_option(
            "--processes", action="store",
            default=env.get('NOSE_PROCESSES', 0),
            dest="multiprocess_workers", metavar="NUM",
            help="Spread test run among this many processes. "
            "Set a number equal to the number of processors "
            "or cores in your machine for best results. "
            "Pass a negative number to have the number of "
            "processes automatically set to the number of cores. "
            "Passing 0 means to disable parallel testing. "
            "Default is 0 unless NOSE_PROCESSES is set. [NOSE_PROCESSES]")
        parser.add_option(
            "--process-timeout", action="store",
            default=env.get('NOSE_PROCESS_TIMEOUT', 10),
            dest="multiprocess_timeout", metavar="SECONDS",
            help="Set timeout for return of results from each "
            "test runner process. Default is 10. [NOSE_PROCESS_TIMEOUT]")
        parser.add_option(
            "--process-restartworker", action="store_true",
            default=env.get('NOSE_PROCESS_RESTARTWORKER', False),
            dest="multiprocess_restartworker",
            help="If set, will restart each worker process once"
            " their tests are done, this helps control memory "
            "leaks from killing the system. [NOSE_PROCESS_RESTARTWORKER]")

    def configure(self, options, config):
        """Configure plugin."""
        try:
            self.status.pop('active')
        except KeyError:
            pass
        if not hasattr(options, 'multiprocess_workers'):
            self.enabled = False
            return
        if config.worker:
            return  # Don't start inside of a worker process
        self.config = config
        try:
            workers = int(options.multiprocess_workers)
        except (TypeError, ValueError):
            workers = 0
        if workers:
            _import_mp()
            if Process is None:
                self.enabled = False
                return
            if workers < 0:
                import multiprocessing
                workers = multiprocessing.cpu_count()
            self.enabled = True
            self.config.multiprocess_workers = workers
            t = float(options.multiprocess_timeout)
            self.config.multiprocess_timeout = t
            r = int(options.multiprocess_restartworker)
            self.config.multiprocess_restartworker = r
            self.status['active'] = True

    def prepareTestLoader(self, loader):
        """Remember loader class so MultiProcessTestRunner can instantiate
        the right loader."""
        self.loaderClass = loader.__class__

    def prepareTestRunner(self, runner):
        """Replace test runner with MultiProcessTestRunner."""
        return MultiProcessTestRunner(
            stream=runner.stream, verbosity=self.config.verbosity,
            config=self.config, loaderClass=self.loaderClass)


def signalhandler(sig, frame):
    raise TimedOutException()


class MultiProcessTestRunner(TextTestRunner):
    # Max time to wait to terminate a process that does not respond to SIGILL.
    waitkilltime = 5.0

    def __init__(self, **kw):
        self.loaderClass = kw.pop('loaderClass', loader.defaultTestLoader)
        super(MultiProcessTestRunner, self).__init__(**kw)

    def collect(self, test, testQueue, tasks, to_teardown, result):
        for case in self.nextBatch(test):
            log.debug("Next batch %s (%s)", case, type(case))
            if (
                isinstance(case, nose.case.Test)
                and isinstance(case.test, failure.Failure)
            ):
                log.debug("Case is a Failure")
                case(result)
                continue
            if isinstance(
                case, ContextSuite
            ) and case.context is failure.Failure:
                log.debug("Case is a Failure")
                case(result)
                continue
            elif isinstance(case, ContextSuite) and self.sharedFixtures(case):
                log.debug("%s has shared fixtures", case)
                try:
                    case.setUp()
                except (KeyboardInterrupt, SystemExit):
                    raise
                except Exception:
                    log.debug("%s setup failed", sys.exc_info())
                    result.addError(case, sys.exc_info())
                else:
                    to_teardown.append(case)
                    if case.factory:
                        ancestors = case.factory.context.get(case, [])
                        for an in ancestors[:2]:
                            if getattr(an, '_multiprocess_shared_', False):
                                an._multiprocess_can_split_ = True
                    self.collect(case, testQueue, tasks, to_teardown, result)
            else:
                test_addr = self.addtask(testQueue, tasks, case)
                log.debug(
                    "Queued test %s (%s) to %s",
                    len(tasks), test_addr, testQueue,
                )

    def startProcess(
        self, iworker, testQueue, resultQueue, shouldStop, result
    ):
        currentaddr = Value('c', bytes_(''))
        currentstart = Value('d', time.time())
        keyboardCaught = Event()
        p = Process(
            target=runner,
            args=(
                iworker,
                testQueue,
                resultQueue,
                currentaddr,
                currentstart,
                keyboardCaught,
                shouldStop,
                self.loaderClass,
                result.__class__,
                pickle.dumps(self.config),
            )
        )
        p.currentaddr = currentaddr
        p.currentstart = currentstart
        p.keyboardCaught = keyboardCaught
        old = signal.signal(signal.SIGILL, signalhandler)
        p.start()
        signal.signal(signal.SIGILL, old)
        return p

    def run(self, test):
        """Execute the test (which may be a suite). If the test is a suite,
        distribute it out among as many processes as have been configured, at
        as fine a level as is possible given the context fixtures defined in
        the suite or any sub-suites."""
        log.debug("%s.run(%s) (%s)", self, test, os.getpid())
        wrapper = self.config.plugins.prepareTest(test)
        if wrapper is not None:
            test = wrapper
        wrapped = self.config.plugins.setOutputStream(self.stream)
        if wrapped is not None:
            self.stream = wrapped
        testQueue = Queue()
        resultQueue = Queue()
        tasks = []
        completed = []
        workers = []
        to_teardown = []
        shouldStop = Event()
        result = self._makeResult()
        start = time.time()
        self.collect(test, testQueue, tasks, to_teardown, result)
        log.debug("Starting %s workers", self.config.multiprocess_workers)
        for i in range(self.config.multiprocess_workers):
            p = self.startProcess(
                i, testQueue, resultQueue, shouldStop, result
            )
            workers.append(p)
            log.debug("Started worker process %s", i+1)
        total_tasks = len(tasks)
        nexttimeout = self.config.multiprocess_timeout
        thrownError = None
        try:
            while tasks:
                log.debug(
                    "Waiting for results (%s/%s tasks), next timeout=%.3fs",
                    len(completed), total_tasks, nexttimeout,
                )
                try:
                    iworker, addr, newtask_addrs, batch_result = (
                        resultQueue.get(timeout=nexttimeout)
                    )
                    log.debug(
                        'Results received for worker %d, %s, new tasks: %d',
                        iworker, addr, len(newtask_addrs),
                    )
                    try:
                        try:
                            tasks.remove(addr)
                        except ValueError:
                            log.warning(
                                'worker %s failed to remove from tasks: %s',
                                iworker, addr,
                            )
                        total_tasks += len(newtask_addrs)
                        tasks.extend(newtask_addrs)
                    except KeyError:
                        log.debug("Got result for unknown task? %s", addr)
                        log.debug("current: %s", str(list(tasks)[0]))
                    else:
                        completed.append([addr, batch_result])
                    self.consolidate(result, batch_result)
                    if (
                        self.config.stopOnError and not result.wasSuccessful()
                    ):
                        shouldStop.set()
                        break
                    if self.config.multiprocess_restartworker:
                        log.debug('joining worker %s', iworker)
                        workers[iworker].join(timeout=1)
                        if not shouldStop.is_set() and not testQueue.empty():
                            log.debug(
                                'starting new process on worker %s', iworker
                            )
                            workers[iworker] = self.startProcess(
                                iworker,
                                testQueue,
                                resultQueue,
                                shouldStop,
                                result,
                            )
                except Empty:
                    msg = "Timeout - %s tasks pending (empty testQueue=%r): %s"
                    log.debug(msg, len(tasks), testQueue.empty(), str(tasks))
                    any_alive = False
                    for iworker, w in enumerate(workers):
                        if w.is_alive():
                            worker_addr = bytes_(w.currentaddr.value, 'ascii')
                            timeprocessing = time.time() - w.currentstart.value
                            if (
                                len(worker_addr) == 0
                                and timeprocessing
                                > self.config.multiprocess_timeout - 0.1
                            ):
                                log.debug(
                                    'Worker %d finished its work item, but is '
                                    'not exiting. Do we wait for it?', iworker,
                                )
                            else:
                                any_alive = True
                            if (
                                len(worker_addr) > 0
                                and timeprocessing
                                > self.config.multiprocess_timeout - 0.1
                            ):
                                msg = 'Timed out worker %s: %s'
                                log.debug(msg, iworker, worker_addr)
                                w.currentaddr.value = bytes_('')
                                w.keyboardCaught.clear()
                                startkilltime = time.time()
                                while (
                                    not w.keyboardCaught.is_set()
                                    and w.is_alive()
                                ):
                                    if (
                                        time.time() - startkilltime
                                        > self.waitkilltime
                                    ):
                                        msg = "terminating worker %s"
                                        log.error(msg, iworker)
                                        w.terminate()
                                        workers[iworker] = w = (
                                            self.startProcess(
                                                iworker,
                                                testQueue,
                                                resultQueue,
                                                shouldStop,
                                                result,
                                            )
                                        )
                                        break
                                    os.kill(w.pid, signal.SIGILL)
                                    time.sleep(0.1)
                    if not any_alive and testQueue.empty():
                        log.debug("All workers are dead")
                        break
                nexttimeout = self.config.multiprocess_timeout
                for w in workers:
                    if w.is_alive() and len(w.currentaddr.value) > 0:
                        timeprocessing = time.time() - w.currentstart.value
                        if timeprocessing <= self.config.multiprocess_timeout:
                            nexttimeout = min(
                                nexttimeout,
                                self.config.multiprocess_timeout
                                - timeprocessing,
                            )
            log.debug(
                "Completed %s tasks (%s remain)", len(completed), len(tasks)
            )
        except (KeyboardInterrupt, SystemExit) as e:
            log.info('parent received ctrl-c when waiting for test results')
            thrownError = e
            result.addError(test, sys.exc_info())
        try:
            for case in to_teardown:
                log.debug("Tearing down shared fixtures for %s", case)
                try:
                    case.tearDown()
                except (KeyboardInterrupt, SystemExit):
                    raise
                except Exception:
                    result.addError(case, sys.exc_info())
            stop = time.time()
            result.printErrors()
            result.printSummary(start, stop)
            self.config.plugins.finalize(result)
            if thrownError is None:
                log.debug("Tell all workers to stop")
                for w in workers:
                    if w.is_alive():
                        testQueue.put('STOP', block=False)
            for iworker, worker in enumerate(workers):
                if worker.is_alive():
                    log.debug('joining worker %s', iworker)
                    worker.join()
                    if worker.is_alive():
                        log.debug('failed to join worker %s', iworker)
        except (KeyboardInterrupt, SystemExit):
            log.info(
                'parent received ctrl-c when shutting down: stop all processes'
            )
            for worker in workers:
                if worker.is_alive():
                    worker.terminate()
            if thrownError:
                raise thrownError
            else:
                raise
        return result

    def addtask(testQueue, tasks, case):
        arg = None
        if isinstance(case, nose.case.Test) and hasattr(case.test, 'arg'):
            case.test.descriptor = None
            arg = case.test.arg
        test_addr = MultiProcessTestRunner.address(case)
        testQueue.put((test_addr, arg), block=False)
        if arg is not None:
            test_addr += str(arg)
        if tasks is not None:
            tasks.append(test_addr)
        return test_addr
    addtask = staticmethod(addtask)

    def address(case):
        if hasattr(case, 'address'):
            file, mod, call = case.address()
        elif hasattr(case, 'context'):
            file, mod, call = test_address(case.context)
        else:
            raise Exception("Unable to convert %s to address" % case)
        parts = []
        if file is None:
            if mod is None:
                raise Exception("Unaddressable case %s" % case)
            else:
                parts.append(mod)
        else:
            dirname, basename = os.path.split(file)
            if basename.startswith('__init__'):
                file = dirname
            parts.append(file)
        if call is not None:
            parts.append(call)
        return ':'.join(map(str, parts))
    address = staticmethod(address)

    def nextBatch(self, test):
        # Tests can mark themselves as not safe for multiprocess execution.
        if hasattr(test, 'context'):
            if not getattr(test.context, '_multiprocess_', True):
                return
        if (
            (
                isinstance(test, ContextSuite)
                and test.hasFixtures(self.checkCanSplit)
            )
            or not getattr(test, 'can_split', True)
            or not isinstance(test, unittest.TestSuite)
        ):
            if isinstance(test, ContextSuite):
                contained = list(test)
                if (
                    len(contained) == 1
                    and getattr(contained[0], 'context', None) == test.context
                ):
                    test = contained[0]
            yield test
        else:
            for case in test:
                yield from self.nextBatch(case)

    def checkCanSplit(context, fixt):
        """Callback to check whether the fixtures found in a context or
        ancestor are ones we care about. Contexts can tell us that their
        fixtures are reentrant by setting _multiprocess_can_split_."""
        if not fixt:
            return False
        if getattr(context, '_multiprocess_can_split_', False):
            return False
        return True
    checkCanSplit = staticmethod(checkCanSplit)

    def sharedFixtures(self, case):
        context = getattr(case, 'context', None)
        if not context:
            return False
        return getattr(context, '_multiprocess_shared_', False)

    def consolidate(self, result, batch_result):
        log.debug("batch result is %s", batch_result)
        try:
            output, testsRun, failures, errors, errorClasses = batch_result
        except ValueError:
            log.debug("result in unexpected format %s", batch_result)
            failure.Failure(*sys.exc_info())(result)
            return
        self.stream.write(output)
        result.testsRun += testsRun
        result.failures.extend(failures)
        result.errors.extend(errors)
        for key, (storage, label, isfail) in list(errorClasses.items()):
            if key not in result.errorClasses:
                result.errorClasses[key] = ([], label, isfail)
            mystorage, _junk, _junk = result.errorClasses[key]
            mystorage.extend(storage)
        log.debug("Ran %s tests (total: %s)", testsRun, result.testsRun)


def runner(
    ix, testQueue, resultQueue, currentaddr, currentstart,
    keyboardCaught, shouldStop, loaderClass, resultClass, config
):
    try:
        try:
            return __runner(
                ix, testQueue, resultQueue, currentaddr, currentstart,
                keyboardCaught, shouldStop, loaderClass, resultClass, config
            )
        except KeyboardInterrupt:
            log.debug('Worker %s keyboard interrupt, stopping', ix)
    except Empty:
        log.debug("Worker %s timed out waiting for tasks", ix)


def __runner(
    ix, testQueue, resultQueue, currentaddr, currentstart,
    keyboardCaught, shouldStop, loaderClass, resultClass, config
):
    config = pickle.loads(config)
    dummy_parser = config.parserClass()
    if _instantiate_plugins is not None:
        for pluginclass in _instantiate_plugins:
            plugin = pluginclass()
            plugin.addOptions(dummy_parser, {})
            config.plugins.addPlugin(plugin)
    config.plugins.configure(config.options, config)
    config.plugins.begin()
    log.debug("Worker %s executing, pid=%d", ix, os.getpid())
    loader = loaderClass(config=config)
    loader.suiteClass.suiteClass = NoSharedFixtureContextSuite

    def get():
        return testQueue.get(timeout=config.multiprocess_timeout)

    def makeResult():
        stream = _WritelnDecorator(StringIO())
        result = resultClass(
            stream, descriptions=1, verbosity=config.verbosity, config=config
        )
        plug_result = config.plugins.prepareTestResult(result)
        if plug_result:
            return plug_result
        return result

    def batch(result):
        failures = [(TestLet(c), err) for c, err in result.failures]
        errors = [(TestLet(c), err) for c, err in result.errors]
        errorClasses = {}
        for key, (storage, label, isfail) in list(result.errorClasses.items()):
            errorClasses[key] = (
                [(TestLet(c), err) for c, err in storage], label, isfail
            )
        return (
            result.stream.getvalue(),
            result.testsRun,
            failures,
            errors,
            errorClasses
        )
    for test_addr, arg in iter(get, 'STOP'):
        if shouldStop.is_set():
            log.exception('Worker %d STOPPED', ix)
            break
        result = makeResult()
        test = loader.loadTestsFromNames([test_addr])
        test.testQueue = testQueue
        test.tasks = []
        test.arg = arg
        log.debug("Worker %s Test is %s (%s)", ix, test_addr, test)
        try:
            if arg is not None:
                test_addr = test_addr + str(arg)
            currentaddr.value = bytes_(test_addr)
            currentstart.value = time.time()
            test(result)
            currentaddr.value = bytes_('')
            resultQueue.put((ix, test_addr, test.tasks, batch(result)))
        except KeyboardInterrupt as e:
            timeout = isinstance(e, TimedOutException)
            if timeout:
                keyboardCaught.set()
            if len(currentaddr.value):
                if timeout:
                    msg = 'Worker %s timed out, failing current test %s'
                else:
                    msg = (
                        'Worker %s keyboard interrupt, failing current test %s'
                    )
                log.exception(msg, ix, test_addr)
                currentaddr.value = bytes_('')
                failure.Failure(*sys.exc_info())(result)
                resultQueue.put((ix, test_addr, test.tasks, batch(result)))
            else:
                if timeout:
                    msg = 'Worker %s test %s timed out'
                else:
                    msg = 'Worker %s test %s keyboard interrupt'
                log.debug(msg, ix, test_addr)
                resultQueue.put((ix, test_addr, test.tasks, batch(result)))
            if not timeout:
                raise
        except SystemExit:
            currentaddr.value = bytes_('')
            log.exception('Worker %s system exit', ix)
            raise
        except Exception:
            currentaddr.value = bytes_('')
            log.exception(
                "Worker %s error running test or returning results", ix
            )
            failure.Failure(*sys.exc_info())(result)
            resultQueue.put((ix, test_addr, test.tasks, batch(result)))
        if config.multiprocess_restartworker:
            break
    log.debug("Worker %s ending", ix)


class NoSharedFixtureContextSuite(ContextSuite):
    """Context suite that never fires shared fixtures.
    When a context sets _multiprocess_shared_, fixtures in that context
    are executed by the main process. Using this class prevents them
    from executing in the runner process as well."""
    testQueue = None
    tasks = None
    arg = None

    def setupContext(self, context):
        if getattr(context, '_multiprocess_shared_', False):
            return
        super(NoSharedFixtureContextSuite, self).setupContext(context)

    def teardownContext(self, context):
        if getattr(context, '_multiprocess_shared_', False):
            return
        super(NoSharedFixtureContextSuite, self).teardownContext(context)

    def run(self, result):
        """Run tests in suite inside of suite fixtures."""
        log.debug(
            "suite %s (%s) run called, tests: %s", id(self), self, self._tests
        )
        if self.resultProxy:
            result, orig = self.resultProxy(result, self), result
        else:
            result, orig = result, result
        try:
            # log.debug('setUp for %s', id(self));
            self.setUp()
        except KeyboardInterrupt:
            raise
        except Exception:
            self.error_context = 'setup'
            result.addError(self, self._exc_info())
            return
        try:
            for test in self._tests:
                if isinstance(test, nose.case.Test) and self.arg is not None:
                    test.test.arg = self.arg
                else:
                    test.arg = self.arg
                test.testQueue = self.testQueue
                test.tasks = self.tasks
                if result.shouldStop:
                    log.debug("stopping")
                    break
                try:
                    test(orig)
                except KeyboardInterrupt as e:
                    timeout = isinstance(e, TimedOutException)
                    if timeout:
                        msg = 'Timeout when running test %s in suite %s'
                    else:
                        msg = (
                            'KeyboardInterrupt '
                            'when running test %s in suite %s'
                        )
                    log.debug(msg, test, self)
                    err = (
                        TimedOutException,
                        TimedOutException(str(test)),
                        sys.exc_info()[2],
                    )
                    test.config.plugins.addError(test, err)
                    orig.addError(test, err)
                    if not timeout:
                        raise
        finally:
            self.has_run = True
            try:
                # log.debug('tearDown for %s', id(self))
                self.tearDown()
            except KeyboardInterrupt:
                raise
            except Exception:
                self.error_context = 'teardown'
                result.addError(self, self._exc_info())



"""Use the AllModules plugin by passing ``--all-modules`` or setting the
NOSE_ALL_MODULES environment variable to enable collection and execution of
tests in all python modules. Normal nose behavior is to look for tests only in
modules that match testMatch."""
from nose.plugins.base import Plugin


class AllModules(Plugin):
    """Collect tests from all python modules."""
    def options(self, parser, env):
        """Register commandline options."""
        env_opt = 'NOSE_ALL_MODULES'
        parser.add_option(
            '--all-modules', action="store_true",
            dest=self.enableOpt, default=env.get(env_opt),
            help="Enable plugin %s: %s [%s]" %
            (self.__class__.__name__, self.help(), env_opt)
        )

    def wantFile(self, file):
        """Override to return True for all files ending with .py"""
        if file.endswith('.py'):
            return True
        return None

    def wantModule(self, module):
        """Override return True for all modules"""
        return True



"""Lists builtin plugins."""
plugins = []
builtins = (
    ('nose.plugins.attrib', 'AttributeSelector'),
    ('nose.plugins.capture', 'Capture'),
    ('nose.plugins.logcapture', 'LogCapture'),
    ('nose.plugins.cover', 'Coverage'),
    ('nose.plugins.debug', 'Pdb'),
    ('nose.plugins.deprecated', 'Deprecated'),
    ('nose.plugins.doctests', 'Doctest'),
    ('nose.plugins.isolate', 'IsolationPlugin'),
    ('nose.plugins.failuredetail', 'FailureDetail'),
    ('nose.plugins.prof', 'Profile'),
    ('nose.plugins.skip', 'Skip'),
    ('nose.plugins.testid', 'TestId'),
    ('nose.plugins.multiprocess', 'MultiProcess'),
    ('nose.plugins.xunit', 'Xunit'),
    ('nose.plugins.allmodules', 'AllModules'),
    ('nose.plugins.collect', 'CollectOnly'),
)

for module, cls in builtins:
    try:
        plugmod = __import__(module, globals(), locals(), [cls])
    except KeyboardInterrupt:
        raise
    except Exception:
        continue
    plug = getattr(plugmod, cls)
    plugins.append(plug)
    globals()[cls] = plug



"""This plugin captures stdout during test execution. If a test fails
or raises an error, the captured output will be appended to the error
or failure output. It is disabled by default, but can be enabled with
the options: ``--capture-output`` or ``--capture_output``.
Or enable it by setting os.environ["NOSE_CAPTURE"] to "1".

:Options:
  ``--capture-output`` or ``--capture_output``
    Capture stdout (stdout output will not be printed) """
import logging
import os
import sys
from nose.plugins.base import Plugin
from nose.pyversion import exc_to_unicode, force_unicode
from nose.util import ln
from io import StringIO

log = logging.getLogger(__name__)


class Capture(Plugin):
    """Output-capturing plugin. Now disabled by default.
    Can be enabled with ``--capture-output`` or ``--capture_output``.
    Or enable it with os.environ["NOSE_CAPTURE"]="1" before tests."""
    enabled = True
    name = "capture"
    score = 1600

    def __init__(self):
        self.stdout = []
        self._buf = None

    def options(self, parser, env):
        """Register commandline options"""
        parser.add_option(
            "-s", "--nocapture", action="store_false",
            default=False, dest="capture",
            help="Don't capture stdout (any stdout output "
            "will be printed immediately) [NOSE_NOCAPTURE]"
        )
        parser.add_option(
            "--capture-output", "--capture_output", action="store_true",
            default=False, dest="capture_output",
            help="Capture stdout (stdout output "
            "will not be printed) [NOSE_CAPTURE]"
        )

    def configure(self, options, conf):
        """Configure plugin. Plugin is enabled by default."""
        self.conf = conf
        if (
            "NOSE_CAPTURE" in os.environ
            and os.environ["NOSE_CAPTURE"] == "1"
        ) or options.capture_output:
            self.enabled = True
        elif not options.capture:
            self.enabled = False

    def afterTest(self, test):
        """Clear capture buffer."""
        self.end()
        self._buf = None

    def begin(self):
        """Replace sys.stdout with capture buffer."""
        self.start()  # get an early handle on sys.stdout

    def beforeTest(self, test):
        """Flush capture buffer."""
        self.start()

    def formatError(self, test, err):
        """Add captured output to error report."""
        test.capturedOutput = output = self.buffer
        self._buf = None
        if not output:
            # Don't return None as that will prevent other
            # formatters from formatting and remove earlier formatters
            # formats, instead return the err we got
            return err
        ec, ev, tb = err
        return (ec, self.addCaptureToErr(ev, output), tb)

    def formatFailure(self, test, err):
        """Add captured output to failure report."""
        return self.formatError(test, err)

    def addCaptureToErr(self, ev, output):
        ev = exc_to_unicode(ev)
        output = force_unicode(output)
        return '\n'.join(
            [
                ev,
                ln('>> begin captured stdout <<'),
                output,
                ln('>> end captured stdout <<')
            ]
        )

    def start(self):
        self.stdout.append(sys.stdout)
        self._buf = StringIO()
        sys.stdout = self._buf

    def end(self):
        if self.stdout:
            sys.stdout = self.stdout.pop()

    def finalize(self, result):
        """Restore stdout."""
        while self.stdout:
            self.end()

    def _get_buffer(self):
        if self._buf is not None:
            return self._buf.getvalue()
        return None

    buffer = property(_get_buffer, None, None, """Captured stdout output.""")



import sys
from nose.core import collector, main, run, run_exit, runmodule
from nose.exc import SkipTest, DeprecatedTest
from nose.tools import with_setup
from nose.__version__ import __version__
import collections

__author__ = 'Jason Pellerin (Original) / Michael Mintz (Update)'
version_list = [int(i) for i in __version__.split(".") if i.isdigit()]
__versioninfo__ = tuple(version_list)
__version__ = '.'.join(map(str, __versioninfo__))
__all__ = [
    'main', 'run', 'run_exit', 'runmodule', 'with_setup',
    'SkipTest', 'DeprecatedTest', 'collector'
]
if sys.version_info >= (3, 10):
    collections.Callable = collections.abc.Callable



"""
Test Result
-----------
Provides a TextTestResult that extends unittest's _TextTestResult to
provide support for error classes (such as the builtin skip and deprecated
classes), and hooks for plugins to take over or extend reporting."""
import logging
from unittest import TextTestResult as _TextTestResult
from nose.config import Config
from nose.util import isclass, ln as _ln

log = logging.getLogger('nose.result')


def _exception_detail(exc):
    try:
        return str(exc)
    except Exception:
        return '<unprintable %s object>' % type(exc).__name__


class TextTestResult(_TextTestResult):
    """Text test result that extends unittest's default test result
    support for a configurable set of errorClasses (eg, Skip,
    Deprecated, TODO) that extend the errors/failures/success triad."""
    def __init__(self, stream, descriptions, verbosity, config=None,
                 errorClasses=None):
        if errorClasses is None:
            errorClasses = {}
        self.errorClasses = errorClasses
        if config is None:
            config = Config()
        self.config = config
        _TextTestResult.__init__(self, stream, descriptions, verbosity)
        _TextTestResult.addDuration = None

    def addSkip(self, test, reason):
        from nose.plugins.skip import SkipTest
        if SkipTest in self.errorClasses:
            storage, label, isfail = self.errorClasses[SkipTest]
            storage.append((test, reason))
            self.printLabel(label, (SkipTest, reason, None))

    def addError(self, test, err):
        """Overrides normal addError to add support for errorClasses."""
        ec, ev, tb = err
        try:
            exc_info = self._exc_info_to_string(err, test)
        except TypeError:
            exc_info = self._exc_info_to_string(err)
        for cls, (storage, label, isfail) in list(self.errorClasses.items()):
            if isclass(ec) and issubclass(ec, cls):
                if isfail:
                    test.passed = False
                storage.append((test, exc_info))
                self.printLabel(label, err)
                return
        self.errors.append((test, exc_info))
        test.passed = False
        self.printLabel('ERROR')

    def getDescription(self, test):
        if self.descriptions:
            return test.shortDescription() or str(test)
        else:
            return str(test)

    def printLabel(self, label, err=None):
        stream = getattr(self, 'stream', None)
        if stream is not None:
            if self.showAll:
                message = [label]
                if err:
                    detail = _exception_detail(err[1])
                    if detail:
                        message.append(detail)
                stream.writeln(": ".join(message))
            elif self.dots:
                stream.write(label[:1])

    def printErrors(self):
        """Overrides to print all errorClasses errors as well."""
        _TextTestResult.printErrors(self)
        for cls in list(self.errorClasses.keys()):
            storage, label, isfail = self.errorClasses[cls]
            if isfail:
                self.printErrorList(label, storage)
        if hasattr(self, 'config'):
            self.config.plugins.report(self.stream)

    def printSummary(self, start, stop):
        """Called by the test runner to print the final results summary."""
        write = self.stream.write
        writeln = self.stream.writeln
        taken = float(stop - start)
        run = self.testsRun
        plural = run != 1 and "s" or ""
        writeln(self.separator2)
        writeln("Ran %s test%s in %.3fs" % (run, plural, taken))
        writeln()
        summary = {}
        eckeys = list(self.errorClasses.keys())
        for cls in eckeys:
            storage, label, isfail = self.errorClasses[cls]
            count = len(storage)
            if not count:
                continue
            summary[label] = count
        if len(self.failures):
            summary['failures'] = len(self.failures)
        if len(self.errors):
            summary['errors'] = len(self.errors)
        if not self.wasSuccessful():
            write("FAILED")
        else:
            write("OK")
        items = list(summary.items())
        if items:
            items.sort()
            write(" (")
            write(", ".join(["%s=%s" % (label, count) for
                             label, count in items]))
            writeln(")")
        else:
            writeln()

    def wasSuccessful(self):
        if self.errors or self.failures:
            return False
        for cls in list(self.errorClasses.keys()):
            storage, label, isfail = self.errorClasses[cls]
            if not isfail:
                continue
            if storage:
                return False
        return True

    def _addError(self, test, err):
        try:
            exc_info = self._exc_info_to_string(err, test)
        except TypeError:
            exc_info = self._exc_info_to_string(err)
        self.errors.append((test, exc_info))
        if self.showAll:
            self.stream.write('ERROR')
        elif self.dots:
            self.stream.write('E')

    def _exc_info_to_string(self, err, test=None):
        from nose.plugins.skip import SkipTest
        if isclass(err[0]) and issubclass(err[0], SkipTest):
            return str(err[1])
        try:
            return _TextTestResult._exc_info_to_string(self, err, test)
        except TypeError:
            return _TextTestResult._exc_info_to_string(self, err)


def ln(*arg, **kw):
    from warnings import warn
    warn("ln() has moved to nose.util from nose.result and will be removed "
         "from nose.result in a future release. Please update your imports ",
         DeprecationWarning)
    return _ln(*arg, **kw)



"""Adds a sphinx directive that can be used to automatically document a plugin.
This:
 .. autoplugin :: nose.plugins.foo
    :plugin: Pluggy
Produces:
  .. automodule :: nose.plugins.foo
Options
-------
  .. cmdoption :: --foo=BAR, --fooble=BAR
    Do the foo thing to the new thing.
Plugin
------
  .. autoclass :: nose.plugins.foo.Pluggy
     :members:
Source
------
  .. include :: path/to/nose/plugins/foo.py
     :literal: """
import os
try:
    from docutils import nodes, utils
    from docutils.statemachine import ViewList
    from docutils.parsers.rst import directives
except ImportError:
    pass
from nose.util import resolve_name
from nose.plugins.base import Plugin
from nose.plugins.manager import BuiltinPluginManager
from nose.config import Config
from nose.core import TestProgram
from inspect import isclass


def autoplugin_directive(dirname, arguments, options, content, lineno,
                         content_offset, block_text, state, state_machine):
    mod_name = arguments[0]
    mod = resolve_name(mod_name)
    plug_name = options.get("plugin", None)
    if plug_name:
        obj = getattr(mod, plug_name)
    else:
        for entry in dir(mod):
            obj = getattr(mod, entry)
            if isclass(obj) and issubclass(obj, Plugin) and obj is not Plugin:
                plug_name = '%s.%s' % (mod_name, entry)
                break
    rst = ViewList()
    rst.append('.. automodule :: %s\n' % mod_name, '<autodoc>')
    rst.append('', '<autodoc>')
    rst.append('Options', '<autodoc>')
    rst.append('-------', '<autodoc>')
    rst.append('', '<autodoc>')
    plug = obj()
    opts = OptBucket()
    plug.options(opts, {})
    for opt in opts:
        rst.append(opt.options(), '<autodoc>')
        rst.append('   \n', '<autodoc>')
        rst.append('   ' + opt.help + '\n', '<autodoc>')
        rst.append('\n', '<autodoc>')
    rst.append('Plugin', '<autodoc>')
    rst.append('------', '<autodoc>')
    rst.append('', '<autodoc>')
    rst.append('.. autoclass :: %s\n' % plug_name, '<autodoc>')
    rst.append('   :members:\n', '<autodoc>')
    rst.append('   :show-inheritance:\n', '<autodoc>')
    rst.append('', '<autodoc>')
    rst.append('Source', '<autodoc>')
    rst.append('------', '<autodoc>')
    rst.append(
        '.. include :: %s\n' % utils.relative_path(
            state_machine.document['source'],
            os.path.abspath(mod.__file__.replace('.pyc', '.py'))),
        '<autodoc>')
    rst.append('   :literal:\n', '<autodoc>')
    rst.append('', '<autodoc>')
    node = nodes.section()
    node.document = state.document
    surrounding_title_styles = state.memo.title_styles
    surrounding_section_level = state.memo.section_level
    state.memo.title_styles = []
    state.memo.section_level = 0
    state.nested_parse(rst, 0, node, match_titles=1)
    state.memo.title_styles = surrounding_title_styles
    state.memo.section_level = surrounding_section_level
    return node.children


def autohelp_directive(dirname, arguments, options, content, lineno,
                       content_offset, block_text, state, state_machine):
    config = Config(parserClass=OptBucket, plugins=BuiltinPluginManager())
    parser = config.getParser(TestProgram.usage())
    rst = ViewList()
    for line in parser.format_help().split('\n'):
        rst.append(line, '<autodoc>')
    rst.append('Options', '<autodoc>')
    rst.append('-------', '<autodoc>')
    rst.append('', '<autodoc>')
    for opt in parser:
        rst.append(opt.options(), '<autodoc>')
        rst.append('   \n', '<autodoc>')
        rst.append('   ' + opt.help + '\n', '<autodoc>')
        rst.append('\n', '<autodoc>')
    node = nodes.section()
    node.document = state.document
    surrounding_title_styles = state.memo.title_styles
    surrounding_section_level = state.memo.section_level
    state.memo.title_styles = []
    state.memo.section_level = 0
    state.nested_parse(rst, 0, node, match_titles=1)
    state.memo.title_styles = surrounding_title_styles
    state.memo.section_level = surrounding_section_level
    return node.children


class OptBucket(object):
    def __init__(self, doc=None, prog='nosetests'):
        self.opts = []
        self.doc = doc
        self.prog = prog

    def __iter__(self):
        return iter(self.opts)

    def format_help(self):
        return self.doc.replace('%prog', self.prog).replace(':\n', '::\n')

    def add_option(self, *arg, **kw):
        self.opts.append(Opt(*arg, **kw))


class Opt(object):
    def __init__(self, *arg, **kw):
        self.opts = arg
        self.action = kw.pop('action', None)
        self.default = kw.pop('default', None)
        self.metavar = kw.pop('metavar', None)
        self.help = kw.pop('help', None)

    def options(self):
        buf = []
        for optstring in self.opts:
            desc = optstring
            if self.action not in ('store_true', 'store_false'):
                desc += '=%s' % self.meta(optstring)
            buf.append(desc)
        return '.. cmdoption :: ' + ', '.join(buf)

    def meta(self, optstring):
        return self.metavar or 'DEFAULT'


def setup(app):
    app.add_directive('autoplugin',
                      autoplugin_directive, 1, (1, 0, 1),
                      plugin=directives.unchanged)
    app.add_directive('autohelp', autohelp_directive, 0, (0, 0, 1))



"""Simple traceback introspection. Used to add additional information to
AssertionErrors in tests, so that failure messages may be more informative."""
import inspect
import logging
import re
import textwrap
import tokenize
import collections
from io import StringIO

log = logging.getLogger(__name__)


def inspect_traceback(tb):
    """Inspect a traceback and its frame, returning source for the expression
    where the exception was raised, with simple variable replacement performed
    and the line on which the exception was raised marked with '>>' """
    log.debug('inspect traceback %s', tb)
    while tb.tb_next:
        tb = tb.tb_next
    frame = tb.tb_frame
    lines, exc_line = tbsource(tb)
    inspect_lines, mark_line = find_inspectable_lines(lines, exc_line)
    src = StringIO(textwrap.dedent(''.join(inspect_lines)))
    exp = Expander(frame.f_locals, frame.f_globals)
    while inspect_lines:
        try:
            for tok in tokenize.generate_tokens(src.readline):
                exp(*tok)
        except tokenize.TokenError as e:
            log.debug("Tokenizer error: %s", e)
            inspect_lines.pop(0)
            mark_line -= 1
            src = StringIO(textwrap.dedent(''.join(inspect_lines)))
            exp = Expander(frame.f_locals, frame.f_globals)
            continue
        break
    padded = []
    if exp.expanded_source:
        exp_lines = exp.expanded_source.split('\n')
        ep = 0
        for line in exp_lines:
            if ep == mark_line:
                padded.append('>>  ' + line)
            else:
                padded.append('    ' + line)
            ep += 1
    return '\n'.join(padded)


def tbsource(tb, context=6):
    """Get source from  a traceback object.
    A tuple of two things is returned: a list of lines of context from
    the source code, and the index of the current line within that list.
    The optional second argument specifies the number of lines of context
    to return, which are centered around the current line."""
    lineno = tb.tb_lineno
    frame = tb.tb_frame
    if context > 0:
        start = lineno - 1 - context//2
        log.debug("lineno: %s start: %s", lineno, start)
        try:
            lines, dummy = inspect.findsource(frame)
        except IOError:
            lines, index = [''], 0
        else:
            all_lines = lines
            start = max(start, 1)
            start = max(0, min(start, len(lines) - context))
            lines = lines[start:start+context]
            index = lineno - 1 - start
            if index > 0:
                while lines[index-1].strip().endswith('\\'):
                    start -= 1
                    lines = all_lines[start:start+context]
    else:
        lines, index = [''], 0
    log.debug("tbsource lines '''%s''' around index %s", lines, index)
    return (lines, index)


def find_inspectable_lines(lines, pos):
    r"""Find lines in home that are inspectable.
    Walk back from the err line up to 3 lines, but don't walk back over
    changes in indent level.
    Walk forward up to 3 lines, counting \ separated lines as 1. Don't walk
    over changes in indent level (unless part of an extended line)"""
    cnt = re.compile(r'\\[\s\n]*$')
    df = re.compile(r':[\s\n]*$')
    ind = re.compile(r'^(\s*)')
    toinspect = []
    home = lines[pos]
    home_indent = ind.match(home).groups()[0]
    before = lines[max(pos-3, 0):pos]
    before.reverse()
    after = lines[pos+1:min(pos+4, len(lines))]
    for line in before:
        if ind.match(line).groups()[0] == home_indent:
            toinspect.append(line)
        else:
            break
    toinspect.reverse()
    toinspect.append(home)
    home_pos = len(toinspect)-1
    continued = cnt.search(home)
    for line in after:
        if (
            (continued or ind.match(line).groups()[0] == home_indent)
            and not df.search(line)
        ):
            toinspect.append(line)
            continued = cnt.search(line)
        else:
            break
    log.debug("Inspecting lines '''%s''' around %s", toinspect, home_pos)
    return toinspect, home_pos


class Expander:
    """Simple expression expander. Uses tokenize to find the names and
    expands any that can be looked up in the frame."""
    def __init__(self, locals, globals):
        self.locals = locals
        self.globals = globals
        self.lpos = None
        self.expanded_source = ''

    def __call__(self, ttype, tok, start, end, line):
        if self.lpos is not None:
            if start[1] >= self.lpos:
                self.expanded_source += ' ' * (start[1]-self.lpos)
            elif start[1] < self.lpos:
                self.expanded_source += ' ' * start[1]
        self.lpos = end[1]
        if ttype == tokenize.INDENT:
            pass
        elif ttype == tokenize.NAME:
            try:
                val = self.locals[tok]
                if isinstance(val, collections.abc.Callable):
                    val = tok
                else:
                    val = repr(val)
            except KeyError:
                try:
                    val = self.globals[tok]
                    if isinstance(val, collections.abc.Callable):
                        val = tok
                    else:
                        val = repr(val)

                except KeyError:
                    val = tok
            self.expanded_source += val
        else:
            self.expanded_source += tok
        if re.match(r'\s+\\\n', line[end[1]:]):
            self.expanded_source += ' \\\n'
